from distutils.core import setup

setup(
    name='tordl',
    packages=['tordl'],
    version='1.04',
    license='MIT',
    description='CLI Torrent Search and Download',
    author='x0r0x',
    author_email='jakub.schimer@protonmail.com',
    url='https://github.com/X0R0X/cli-torrent-dl/',
    download_url='https://github.com/X0R0X/cli-torrent-dl/archive/refs/heads/master.zip',
    keywords=['torrent', 'search', 'download', 'cli', 'curses'],
    install_requires=[
        'aiohttp==3.7.4.post0',
        'async-timeout==3.0.1',
        'attrs==20.3.0',
        'beautifulsoup4==4.9.3',
        'bs4==0.0.1',
        'chardet==4.0.0',
        'idna==3.1',
        'multidict==5.1.0',
        'pkg-resources==0.0.0',
        'soupsieve==2.2.1',
        'typing-extensions==3.7.4.3',
        'yarl==1.6.3',
        'aiohttp==3.7.4.post0',
        'beautifulsoup4',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Other Audience',
        'Intended Audience :: Developers',
        'Topic :: Communications :: File Sharing',
        'Topic :: Home Automation',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Terminals',
        'Environment :: Console',
        'Environment :: Console :: Curses',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],
)
